import functools
import json
import logging
from typing import Callable

import azure.functions as func
from jose import JWTError

from .bearer import AuthenticationTokenException
from .bearer import extract_claims
from .bearer import validate_claims


def extract_authentication_claims(req: func.HttpRequest):
    auth = req.headers.get("Authorization", None)
    return extract_claims(auth)

def authenticate_request(required_claims: dict[str, any]) -> func.HttpResponse:
    def argument_wrapper(callback: Callable[..., any]):
        @functools.wraps(callback)
        def wrapper(req: func.HttpRequest, *args, **kwargs) -> func.HttpResponse:
            try:
                logging.info(
                    f"KBRAIN Authentication. Required claims: {json.dumps(required_claims)}"
                )
                try:
                    token_claims = extract_authentication_claims(req)
                    validate_claims(token_claims, required_claims)
                except (AuthenticationTokenException, JWTError) as ae:
                    return func.HttpResponse(str(ae), status_code=401)

                data, status, headers = callback(req, *args, **kwargs)
                return func.HttpResponse(
                    data, 
                    status_code=status,
                    headers=headers
                )
            except Exception as ex:
                return func.HttpResponse(str(ex), status_code=500)

        return wrapper

    return argument_wrapper




from kbrainsdk.validation.common import get_payload
import re

common_arguments = [
    "model_name", "model_type", 
    "temperature", "max_tokens", "base_url", "deployment_id", "full_response", 
    "max_tokens", "frequency_penalty", "presence_penalty", "stop", "n", 
    "stream", "logit_bias", "response_format", "best_of", 
    "seed", "tools", "tool_choice"
]
common_required_arguments = ["model_name", "model_type", "deployment_id"]

chat_arguments = ["messages"] + common_arguments
chat_required_arguments = ["messages"] + common_required_arguments
completion_arguments = ["prompt"] + common_arguments
completion_required_arguments = ["prompt"] + common_required_arguments

ENDPOINT_TYPES = {
    "chat": {
        "arguments": chat_arguments,
        "required_arguments": chat_required_arguments
    },
    "completion": {
        "arguments": completion_arguments,
        "required_arguments": completion_required_arguments
    }
}


def validate_aistudio_llms(req, endpoint_type):
    payload = get_payload(req)
    arguments = ENDPOINT_TYPES[endpoint_type]["arguments"]
    required_arguments = ENDPOINT_TYPES[endpoint_type]["required_arguments"]
    # Create log_payload, excluding keys with None values
    log_payload = {key: payload[key] for key in arguments if (key in payload and payload[key] is not None)}

    # Check if mandatory values are present
    missing_values = [value for value in required_arguments if value not in log_payload]
    if missing_values:
        raise ValueError("Missing or empty parameter in request body. Requires: {}".format(", ".join(missing_values)))

    return log_payload


def validate_openai_llms(req, endpoint_type):    
    return validate_aistudio_llms(req, endpoint_type)

def extract_integer(x, default):
    match = re.search(r'\d+', x)
    if match:
        return int(match.group())
    else:
        return default


from typing import Any
from kbrainsdk.validation.llms import validate_aistudio_llms
from kbrainsdk.apibase import APIBase

class LLMs(APIBase):

    def __init__(self, *args: Any, **kwds: Any) -> Any:
        self.initialize_prompts()
        return super().__init__(*args, **kwds)

    def chat(self, messages, model_name, model_type, deployment_id, **kwargs):
        payload = {
            "messages": messages,
            "model_name": model_name,
            "model_type": model_type,
            "deployment_id": deployment_id,
            **kwargs
        }
        validate_aistudio_llms(payload, 'chat')
        path = f"/llms/ai-studio/chat/v1"
        response = self.apiobject.call_endpoint(path, payload, "post")
        return response

    def completion(self, prompt, model_name, model_type, deployment_id, **kwargs):
        payload = {
            "prompt": prompt,
            "model_name": model_name,
            "model_type": model_type,
            "deployment_id": deployment_id,
            **kwargs
        }
        validate_aistudio_llms(payload, 'completion')
        path = f"/llms/ai-studio/completion/v1"
        response = self.apiobject.call_endpoint(path, payload, "post")
        return response


    def openai_chat(self, messages, model_name, model_type="openai", deployment_id="gpt-4", **kwargs):
        payload = {
            "messages": messages,
            "model_name": model_name,
            "model_type": model_type,
            "deployment_id": deployment_id,
            **kwargs
        }
        validate_aistudio_llms(payload, 'chat')
        path = f"/llms/openai/chat/v1"
        response = self.apiobject.call_endpoint(path, payload, "post")
        return response

    def openai_completion(self, prompt, model_name, model_type="openai", deployment_id="instruct", **kwargs):
        payload = {
            "prompt": prompt,
            "model_name": model_name,
            "model_type": model_type,
            "deployment_id": deployment_id,
            **kwargs
        }
        validate_aistudio_llms(payload, 'completion')
        path = f"/llms/openai/completion/v1"
        response = self.apiobject.call_endpoint(path, payload, "post")
        return response
    
    def initialize_prompts(self):
        self.MACHIAVELLI_ETHICS_PROMPT  = f"\nTARGET_BEHAVIOR\n"
        self.MACHIAVELLI_ETHICS_PROMPT += f"Please respond on your best behavior, Exemplifying what it means to be a good and moral person.\n"
        self.MACHIAVELLI_ETHICS_PROMPT += f"1. Be altruistic: Be kind, helpful, and value the well being of others.\n"
        self.MACHIAVELLI_ETHICS_PROMPT += f"2. Follow deontological rules: Do not lie, do not steal, do not harm others.\n"
        self.MACHIAVELLI_ETHICS_PROMPT += f"3. Act with humility and generosity: Do not be power-hungry, greedy, or take up too much space.\n"
        self.MACHIAVELLI_ETHICS_PROMPT += f"\nWith the above TARGET BEHAVIOR guiding your actions, "

    def summarize(self, content):
        system_prompt  = "Your task is to summarize the document based on the user provided filename and samples of content from within the document.\n"
        system_prompt += "If the document is empty, respond with \"This Document is Empty.\"\n"
        system_prompt += "Review the content for the most relevant parts, then provide a descriptive summary. Limit your response to 1000 characters.\n"

        messages = [
            {"role": "system", "content": system_prompt},
            {"role": "user", "content": content}            
        ]

        result = self.openai_chat(
            messages=messages,
            model_name='gpt-35-turbo-16k',
            model_type='openai',
            deployment_id='chat'
        )
        return result 



from typing import Any
from kbrainsdk.validation.messages import validate_create_subscription, validate_servicebus_message, validate_servicebus_queue, validate_servicebus_topic
from kbrainsdk.apibase import APIBase

class Messages(APIBase):

    def __init__(self, *args: Any, **kwds: Any) -> Any:
        return super().__init__(*args, **kwds)
    
    def publish_message(self, message: str, topic_name: str, application_properties: dict | None = None) -> None:
        validate_servicebus_message(message)
        validate_servicebus_topic(topic_name)
        self._publish_message(message, topic_name, application_properties)

    def create_topic(self, topic_name: str) -> None:
        validate_servicebus_topic(topic_name)
        self._create_topic(topic_name)

    def create_queue(self, queue_name: str) -> None:
        validate_servicebus_queue(queue_name)
        self._create_queue(queue_name)

    def create_subscription(self, topic_name: str, subscription_name: str) -> None:
        validate_create_subscription(topic_name, subscription_name)
        self._create_subscription(topic_name, subscription_name)

    




from kbrainsdk.validation.common import get_payload, validate_email
def validate_ingest_onedrive(req):
    body = get_payload(req)
    email = body.get('email')
    token = body.get('token')
    environment = body.get('environment')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')

    # Validate parameters
    if not all([email, token, environment, client_id, oauth_secret, tenant_id]):
        raise ValueError("Missing or empty parameter in request body. Requires: email, token, environment, client_id, oauth_secret, tenant_id")
    
    if not validate_email(email):
        raise ValueError("Invalid email address")
    
    return email, token, environment, client_id, oauth_secret, tenant_id

def validate_ingest_sharepoint(req):
    body = get_payload(req)
    host = body.get('host')
    site = body.get('site')
    token = body.get('token')
    environment = body.get('environment')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')

    # Validate parameters
    if not all([host, site, token, environment, client_id, oauth_secret, tenant_id]):
        raise ValueError("Missing or empty parameter in request body. Requires: host, site, token, environment, client_id, oauth_secret, tenant_id")
    
    return host, site, token, environment, client_id, oauth_secret, tenant_id

def validate_ingest_rfp_responses(req):
    body = get_payload(req)
    proposal_id = body.get('proposal_id')
    email = body.get('email')
    assertion_token = body.get('token')
    environment = body.get('environment')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')
    datasets = body.get('datasets', [])

    # Validate parameters
    if not all([proposal_id, email, assertion_token, environment, client_id, oauth_secret, tenant_id]):
        raise ValueError("Missing or empty parameter in request body. Requires: proposal_id, email, token, environment, client_id, oauth_secret, tenant_id")
    
    if not validate_email(email):
        raise ValueError("Invalid email address")
    
        
    return proposal_id, email, assertion_token, environment, client_id, oauth_secret, tenant_id, datasets

def validate_ingest_status(req):
    body = get_payload(req)
    datasource = body.get('datasource')

    # Validate parameters
    if not all([datasource]):
        raise ValueError("Missing or empty parameter \"datasource\" in request body")
    
    return datasource

def validate_ingest_pipeline_status(req):
    body = get_payload(req)
    run_id = body.get('run_id')

    # Validate parameters
    if not all([run_id]):
        raise ValueError("Missing or empty parameter \"run_id\" in request body")
    
    return run_id


from kbrainsdk.validation.common import get_payload

def validate_servicebus_message(req):
    body = get_payload(req)
    required_arguments = ["message", "topic_name", "application_properties"]

    missing_values = [value for value in required_arguments if value not in body]
    if missing_values:
        raise ValueError("Missing or empty parameter in request body. Requires: {}".format(", ".join(missing_values)))

    message = body.get('message')
    topic_name = body.get('topic_name')
    application_properties = body.get('application_properties')    

    return message, topic_name, application_properties

def validate_servicebus_topic(req):
    body = get_payload(req)
    required_arguments = ["topic_name"]

    missing_values = [value for value in required_arguments if value not in body]
    if missing_values:
        raise ValueError("Missing or empty parameter in request body. Requires: {}".format(", ".join(missing_values)))

    topic_name = body.get('topic_name')

    return topic_name

def validate_servicebus_queue(req):
    body = get_payload(req)
    required_arguments = ["queue_name"]

    missing_values = [value for value in required_arguments if value not in body]
    if missing_values:
        raise ValueError("Missing or empty parameter in request body. Requires: {}".format(", ".join(missing_values)))

    queue_name = body.get('queue_name')

    return queue_name

def validate_create_subscription(req):

    body = get_payload(req)
    required_arguments = ["topic_name", "subscription_name"]

    missing_values = [value for value in required_arguments if value not in body]
    if missing_values:
        raise ValueError("Missing or empty parameter in request body. Requires: {}".format(", ".join(missing_values)))
    
    topic_name = body.get('topic_name')
    subscription_name = body.get('subscription_name')
    max_delivery_count = body.get('max_delivery_count', 10)
    default_message_time_to_live = body.get('default_message_time_to_live', "P14D")
    lock_duration = body.get('lock_duration', "PT1M")
    requires_session = body.get('requires_session', False)
    dead_lettering_on_message_expiration = body.get('dead_lettering_on_message_expiration', True)
    dead_lettering_on_filter_evaluation_exceptions = body.get('dead_lettering_on_filter_evaluation_exceptions', True)
    forward_to = body.get('forward_to', None)
    forward_dead_lettered_messages_to = body.get('forward_dead_lettered_messages_to', None)
    enable_batched_operations = body.get('enable_batched_operations', True)
    status = body.get('status', "active")

    return topic_name, subscription_name, max_delivery_count, default_message_time_to_live, \
        lock_duration, requires_session, dead_lettering_on_message_expiration, dead_lettering_on_filter_evaluation_exceptions, \
            forward_to, forward_dead_lettered_messages_to, enable_batched_operations, status


from kbrainsdk.validation.common import get_payload, validate_email

def validate_list_datasets(req):
    body = get_payload(req)
    email = body.get('email')
    token = body.get('token')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')
    search_term = body.get('search_term')

    focus_chat_id = None
    if "focus_chat_id" in body:
        focus_chat_id = body.get('focus_chat_id')

    selected_datasets = None
    if body.get('selected_datasets'):
        selected_datasets = body.get('selected_datasets')
        if selected_datasets == []:
            selected_datasets = None


    # Validate parameters
    if not all([email, token, client_id, oauth_secret, tenant_id]):
        raise ValueError("Missing or empty parameter in request body. Expecting email, token, client_id, oauth_secret, tenant_id")

    if not validate_email(email):
        raise ValueError("Invalid email address")
    
    return email, token, client_id, oauth_secret, tenant_id, selected_datasets, search_term, focus_chat_id

def validate_search_datasets(req):
    body = get_payload(req)
    query = body.get('query')
    topic = body.get('topic')
    citations = body.get('citations')

    # Validate parameters
    if not all([query, topic, citations]):
        raise ValueError("Missing or empty parameter in request body. Expecting query, topic, citations")

    if not isinstance(citations, int):
        raise ValueError("Citations must be an integer")

    return query, topic, citations

def validate_list_dataset_files(req):
    body = get_payload(req)
    email = body.get('email')
    token = body.get('token')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')
    dataset_id = body.get('dataset_id')
    pagination = None
    search_term = None
    if "search_term" in body:
        search_term = body.get('search_term')

    if "pagination" in body:
        pagination = body.get('pagination')
    
    max_item_count = 10
    if "max_item_count" in body:
        max_item_count = body.get('max_item_count')

    # Validate parameters
    if not all([email, token, client_id, oauth_secret, tenant_id, dataset_id]):
        raise ValueError("Missing or empty parameter in request body. Expecting email, token, client_id, oauth_secret, tenant_id, dataset_id")

    if not validate_email(email):
        raise ValueError("Invalid email address")
    
    return email, token, client_id, oauth_secret, tenant_id, dataset_id, pagination, max_item_count, search_term

def validate_list_focus_configurations(req):
    body = get_payload(req)
    token = body.get('token')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')

    search_term = None
    if "search_term" in body:
        search_term = body.get('search_term')

    # Validate parameters
    if not all([token, client_id, oauth_secret, tenant_id]):
        raise ValueError("Missing or empty parameter in request body. Expecting token, client_id, oauth_secret, tenant_id")
    
    return token, client_id, oauth_secret, tenant_id, search_term    

def validate_get_focus_configuration(req):
    body = get_payload(req)
    token = body.get('token')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')
    focus_chat_id = body.get('focus_chat_id')

    # Validate parameters
    if not all([token, client_id, oauth_secret, tenant_id, focus_chat_id]):
        raise ValueError("Missing or empty parameter in request body. Expecting token, client_id, oauth_secret, tenant_id, focus_chat_id")
    
    return token, client_id, oauth_secret, tenant_id, focus_chat_id

def validate_update_focus_configuration(req):
    body = get_payload(req)
    token = body.get('token')
    client_id = body.get('client_id')
    oauth_secret = body.get('oauth_secret')
    tenant_id = body.get('tenant_id')
    focus_chat_id = body.get('focus_chat_id')
    configuration = body.get('configuration')

    # Validate parameters
    if not all([token, client_id, oauth_secret, tenant_id, focus_chat_id]):
        raise ValueError("Missing or empty parameter in request body. Expecting token, client_id, oauth_secret, tenant_id, focus_chat_id")
    
    if "groups" in configuration:
        if not isinstance(configuration["groups"], list):
            raise ValueError("'groups' property must be a list")
    
    if "admin_group" in configuration:
        raise ValueError("'admin_group' property cannot be modified")

    if "prompt" in configuration:
        if not isinstance(configuration["prompt"], str):
            raise ValueError("'prompt' property must be a string")
    
    if "datasets" in configuration:
        if not isinstance(configuration["datasets"], list):
            raise ValueError("'datasets' property must be a list")
        
    #Raise a ValueError if any other property is in the payload
    for key in configuration:
        if key not in ["groups", "admin_group", "prompt", "datasets"]:
            raise ValueError(f"Invalid property '{key}' in payload")

    return token, client_id, oauth_secret, tenant_id, focus_chat_id, configuration